#include "../include/except.h"
#include "../include/vec.h"

#include <limits.h>
#include <stddef.h>
#include <stdlib.h>

const Exception out_of_memory = { "Out Of Memory" }; // OOM Exception.

Vec *vec_init(int flags){
    Vec *v = malloc(sizeof(*v));

    if(v == NULL){
        if(!(flags & SOFTFAIL)) RAISE(out_of_memory);
        return NULL;
    }

    v->items = malloc( sizeof(void *) * VEC_INIT_SZ );
    if(v->items == NULL){
        if(!(flags & SOFTFAIL)) RAISE(out_of_memory);
        return NULL;
    }
    v->len = 0;
    v->cap = VEC_INIT_SZ;
    return v;
}
int vec_realloc(Vec *v, int cap){
    ASSERTED(v);
    ASSERTED(cap < INT_MAX);
    Vec *old = v;
    size_t sz = sizeof(void *) * cap;
    void **items = realloc(v->items, sz);
    if(items){
        v->items = items;
        v->cap = cap;
        return v->cap;
    }
    v = old;
    return -1;
}
void vec_free(Vec *v){
    free(v->items);
    v->len = 0;
    v->cap = 0;
    free(v);
    v = NULL;
}
int vec_set(Vec *v,int index,void *item){
    ASSERTED(v);
    ASSERTED(index < INT_MAX);
    if(index >= 0 && index < v->len){
        v->items[index] = item;
        return index;
    }    
    return -1;
}
void *vec_read(Vec *v,int index){
    ASSERTED(v);
    ASSERTED(index < INT_MAX);
    void *res = NULL;
    if(index >= 0 && index < v->len){
        res = v->items[index];
    }
    return res;
}
int vec_push(Vec *v, void *item){
    ASSERTED(v);
    ASSERTED(item);
    if(v->len < v->cap){
        v->items[v->len++] = item;
        return v->len;
    }
    // We are at capacity, resize
    int status = vec_realloc(v, v->cap * 2);
    if(status > 0){ // If the realloc succeded, add item.
        v->items[v->len++] = item;
        return v->len;
    }
    return -1;
}
void *vec_pop(Vec *v){
    ASSERTED(v);
    void *res = v->items[--v->len];
    v->items[v->len] = NULL;
    return res;
}
void *vec_delete(Vec *v, int index){
    ASSERTED(v);
    ASSERTED(index < INT_MAX);
    int i = 0;
    if(index > v->len) return NULL;
    void *res = v->items[index];
    // Shift elements to the left to fill the gap
    for (i = index; i < v->len - 1; ++i){
        v->items[i] = v->items[i + 1];
    }
    // Set the last element to NULL 
    v->items[v->len - 1] = NULL;
    // Update the total count
    v->len--;
    // Resize the vector if needed
    if (v->cap > VEC_INIT_SZ && v->len <= v->cap / 4){
        vec_realloc(v, v->cap / 2);
    }
    return res;
}


